import abc
import collections
import copy
import enum
import re
import warnings
from typing import Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Union

import wrapt

import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import infer_dtype
from scipy.ndimage import convolve

import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm, colors
from mpl_toolkits.axes_grid1 import make_axes_locatable

from anndata import AnnData
from scanpy.plotting._utils import add_colors_for_categorical_sample_annotation

from cellrank import logging as logg
from cellrank._utils._docs import d
from cellrank._utils._enum import ModeEnum
from cellrank._utils._lineage import Lineage
from cellrank._utils._utils import _densify_squeeze, _minmax, save_fig, valuedispatch
from cellrank.kernels.mixins import IOMixin

__all__ = ["BaseModel"]

_dup_spaces = re.compile(r" +")  # used on repr for underlying model's repr
ArrayLike = Union[np.ndarray, sp.spmatrix, List, Tuple]


class UnknownModelError(RuntimeError):
    pass


class FailedReturnType(ModeEnum):
    PREPARE = enum.auto()
    FIT = enum.auto()
    PREDICT = enum.auto()
    CONFIDENCE_INTERVAL = enum.auto()
    DEFAULT_CONFIDENCE_INTERVAL = enum.auto()
    PLOT = enum.auto()


class ColorType(ModeEnum):
    CONT = enum.auto()
    CAT = enum.auto()
    STR = enum.auto()


def _handle_exception(return_type: FailedReturnType, func: Callable) -> Callable:
    def handle(*, exception_handler: Callable):
        @wrapt.decorator
        def wrapper(wrapped, instance, args, kwargs):
            try:
                if isinstance(instance, FailedModel):
                    instance.reraise()
                return wrapped(*args, **kwargs)
            except Exception as e:  # noqa: BLE001
                return exception_handler(instance, e, *args, **kwargs)

        return wrapper

    def array_output(instance: "BaseModel", exc: BaseException, *_args, **kwargs) -> np.ndarray:
        if not instance._is_bulk:
            raise exc from None

        if kwargs.get("x_test", None) is not None:
            n_obs = kwargs["x_test"].shape[0]
        elif instance.x_test is not None:
            n_obs = instance.x_test.shape[0]
        else:
            n_obs = 200

        return np.full((n_obs, array_shape), np.nan, dtype=instance._dtype)

    def model_output(instance: "BaseModel", exc: BaseException, *_args, **_kwargs) -> "FailedModel":
        if not isinstance(instance, FailedModel):
            instance = FailedModel(instance, exc=exc)
        if not instance._is_bulk:
            instance.reraise()

        return instance

    def no_output(instance: "BaseModel", exc: BaseException, *_args, **_kwargs) -> None:
        if not instance._is_bulk:
            raise exc from None

    @valuedispatch
    def wrapper(mode: FailedReturnType, *_args, **_kwargs):
        raise NotImplementedError(mode)

    @wrapper.register(FailedReturnType.PREPARE)
    @wrapper.register(FailedReturnType.FIT)
    def _(func: Callable) -> Callable:
        return handle(exception_handler=model_output)(func)

    @wrapper.register(FailedReturnType.PREDICT)
    @wrapper.register(FailedReturnType.CONFIDENCE_INTERVAL)
    @wrapper.register(FailedReturnType.DEFAULT_CONFIDENCE_INTERVAL)
    def _(func: Callable) -> Callable:
        return handle(exception_handler=array_output)(func)

    @wrapper.register(FailedReturnType.PLOT)
    def _(func: Callable):
        return handle(exception_handler=no_output)(func)

    return_type = FailedReturnType(return_type)
    array_shape = 1 + (
        return_type
        in (
            FailedReturnType.CONFIDENCE_INTERVAL,
            FailedReturnType.DEFAULT_CONFIDENCE_INTERVAL,
        )
    )

    return wrapper(return_type, func)


class BaseModelMeta(abc.ABCMeta):
    """Metaclass for all base models."""

    def __new__(cls, clsname: str, superclasses: Tuple[type, ...], attributedict: Dict[str, Any]):
        """Create a new instance.

        Parameters
        ----------
        clsname
            Name of class to be constructed.
        superclasses
            List of superclasses.
        attributedict
            Dictionary of attributes.
        """
        obj = super().__new__(cls, clsname, superclasses, attributedict)
        for fun_name in list(FailedReturnType):
            setattr(
                obj,
                fun_name,
                _handle_exception(FailedReturnType(fun_name), getattr(obj, fun_name)),
            )

        return obj


@d.get_sections(base="base_model", sections=["Parameters"])
@d.dedent
class BaseModel(IOMixin, abc.ABC, metaclass=BaseModelMeta):
    """Base class for all model classes.

    Parameters
    ----------
    %(adata)s
    model
        The underlying model that is used for fitting and prediction.
    """

    _dtype = np.float64

    def __init__(
        self,
        adata: Optional[AnnData],
        model: Any,
    ):
        if not isinstance(adata, AnnData) and not isinstance(self, FittedModel):
            # FittedModel doesn't need it
            raise TypeError(f"Expected `adata` to be of type `anndata.AnnData`, found `{type(adata).__name__}`.")
        super().__init__()

        self._adata = adata
        self._n_obs = 0 if adata is None else adata.n_obs
        self._model = model
        self._gene = None
        self._use_raw = False
        self._data_key = None
        self._lineage = None
        self._prepared = False

        self._obs_names = None
        self._is_bulk = False

        self._x_all = None
        self._y_all = None
        self._w_all = None

        self._x = None
        self._y = None
        self._w = None

        self._x_test = None
        self._y_test = None
        self._x_hat = None
        self._y_hat = None

        self._conf_int = None

    @d.get_summary(base="base_model_prepared")
    @property
    def prepared(self):
        """Whether the model is prepared for fitting."""
        return self._prepared

    @property
    @d.dedent
    def adata(self) -> AnnData:
        """Annotated data object."""
        return self._adata

    @adata.setter
    def adata(self, adata: Optional[AnnData]) -> None:
        self._adata = adata

    @property
    def shape(self) -> Tuple[int]:
        """Number of cells in :attr:`adata`."""
        return (self._n_obs,)

    @property
    def model(self) -> Any:
        """Underlying model."""
        return self._model

    @property
    @d.get_summary(base="base_model_x_all")
    def x_all(self) -> np.ndarray:
        """Unfiltered independent variables of shape ``(n_cells, 1)``."""
        return self._x_all

    @property
    @d.get_summary(base="base_model_y_all")
    def y_all(self) -> np.ndarray:
        """Unfiltered dependent variables of shape ``(n_cells, 1)``."""
        return self._y_all

    @property
    @d.get_summary(base="base_model_w_all")
    def w_all(self) -> np.ndarray:
        """Unfiltered weights of shape ``(n_cells,)``."""
        return self._w_all

    @property
    @d.get_summary(base="base_model_x")
    def x(self) -> np.ndarray:
        """Filtered independent variables of shape ``(n_filtered_cells, 1)`` used for fitting."""
        return self._x

    @property
    @d.get_summary(base="base_model_y")
    def y(self) -> np.ndarray:
        """Filtered dependent variables of shape ``(n_filtered_cells, 1)`` used for fitting."""
        return self._y

    @property
    @d.get_summary(base="base_model_w")
    def w(self) -> np.ndarray:
        """Filtered weights of shape ``(n_filtered_cells,)`` used for fitting."""
        return self._w

    @property
    @d.get_summary(base="base_model_x_test")
    def x_test(self) -> np.ndarray:
        """Independent variables of shape ``(n_samples, 1)`` used for prediction."""
        return self._x_test

    @property
    @d.get_summary(base="base_model_y_test")
    def y_test(self) -> np.ndarray:
        """Prediction values of shape ``(n_samples,)`` for :attr:`x_test`."""
        return self._y_test

    @property
    @d.get_summary(base="base_model_x_hat")
    def x_hat(self) -> np.ndarray:
        """Filtered independent variables used when calculating default confidence interval, usually same as :attr:`x`."""  # noqa: E501
        return self._x_hat

    @property
    @d.get_summary(base="base_model_y_hat")
    def y_hat(self) -> np.ndarray:
        """Filtered dependent variables used when calculating default confidence interval, usually same as :attr:`y`."""
        return self._y_hat

    @property
    @d.get_summary(base="base_model_conf_int")
    def conf_int(self) -> np.ndarray:
        """Array of shape ``(n_samples, 2)`` containing the lower and upper bound of the confidence interval."""
        return self._conf_int

    @d.get_sections(base="base_model_prepare", sections=["Parameters", "Returns"])
    @d.get_full_description(base="base_model_prepare")
    @d.dedent
    def prepare(
        self,
        gene: str,
        lineage: Optional[str],
        time_key: str,
        backward: bool = False,
        time_range: Optional[Union[float, Tuple[float, float]]] = None,
        data_key: Optional[str] = "X",
        use_raw: bool = False,
        threshold: Optional[float] = None,
        weight_threshold: Union[float, Tuple[float, float]] = (0.01, 0.01),
        filter_cells: Optional[float] = None,
        n_test_points: int = 200,
    ) -> "BaseModel":
        """Prepare the model to be ready for fitting.

        Parameters
        ----------
        gene
            Gene in :attr:`~anndata.AnnData.var_names`.
        lineage
            Name of the lineage. If :obj:`None`, all weights will be set to :math:`1`.
        time_key
            Key in :attr:`~anndata.AnnData.obs` where the pseudotime is stored.
        %(backward)s
        %(time_range)s
        data_key
            Key in :attr:`~anndata.AnnData.layers` or ``'X'`` for :attr:`~anndata.AnnData.X`.
            If ``use_raw = True``, it's always set to ``'X'``.
        use_raw
            Whether to access :attr:`~anndata.AnnData.raw`.
        threshold
            Consider only cells with weights > ``threshold`` when estimating the test endpoint.
            If :obj:`None`, use the median of the weights.
        weight_threshold
            Set all weights below ``weight_threshold`` to ``weight_threshold`` if a :class:`float`,
            or to the second value, if a :class:`tuple`.
        filter_cells
            Filter out all cells with expression values lower than this threshold.
        n_test_points
            Number of test points. If :obj:`None`, use the original points based on ``threshold``.

        Returns
        -------
        Nothing, just updates the following fields:

        - :attr:`x` - %(base_model_x.summary)s
        - :attr:`y` - %(base_model_y.summary)s
        - :attr:`w` - %(base_model_w.summary)s
        - :attr:`x_all` - %(base_model_x_all.summary)s
        - :attr:`y_all` - %(base_model_y_all.summary)s
        - :attr:`w_all` - %(base_model_w_all.summary)s
        - :attr:`x_test` - %(base_model_x_test.summary)s
        - :attr:`prepared` - %(base_model_prepared.summary)s
        """
        if use_raw:
            if self.adata.raw is None:
                raise AttributeError("AnnData object has no attribute `.raw`.")
            if data_key != "X":
                data_key = "X"
        self._use_raw = use_raw

        if data_key not in ["X", "obs", None] + list(self.adata.layers.keys()):
            raise KeyError(
                f"Data key must be a key of `adata.layers`: `{list(self.adata.layers.keys())}`, "
                f"`adata.X` or `adata.obs`."
            )

        probs = Lineage.from_adata(self.adata, backward=backward)
        if lineage is not None:
            probs = probs[lineage]

        if time_key not in self.adata.obs:
            raise KeyError(f"Time key `{time_key!r}` not found in `adata.obs`.")

        if data_key == "obs":
            if gene not in self.adata.obs:
                raise KeyError(f"Unable to find data in `adata.obs[{gene!r}]`.")
        else:
            if use_raw and gene not in self.adata.raw.var_names:
                raise KeyError(f"Gene `{gene!r}` not found in `adata.raw.var_names`.")
            if not use_raw and gene not in self.adata.var_names:
                raise KeyError(f"Gene `{gene!r}` not found in `adata.var_names`.")

        if not isinstance(time_range, (type(None), float, int, tuple)):
            raise TypeError(
                f"Expected time range to be either `None`, "
                f"`float` or `tuple`, found `{type(time_range).__name__!r}`."
            )
        if isinstance(time_range, tuple):
            if len(time_range) != 2:
                raise ValueError(f"Expected time range to be a `tuple` of length `2`, found `{len(time_range)}`.")
            if not isinstance(time_range[0], (float, int, type(None))) or not isinstance(
                time_range[1], (float, int, type(None))
            ):
                raise TypeError(
                    f"Expected values in time ranges be of types `float` or `int` or `None`, "
                    f"got `{type(time_range[0]).__name__!r}` and `{type(time_range[1]).__name__!r}`."
                )
            val_start, val_end = time_range
        elif time_range is None:
            val_start, val_end = None, None
        else:
            val_start, val_end = None, time_range

        if isinstance(weight_threshold, (int, float, np.number)):
            weight_threshold = (weight_threshold, weight_threshold)
        if len(weight_threshold) != 2:
            raise ValueError(f"Expected `weight_threshold` to be of size `2`, found `{len(weight_threshold)}`.")

        self._obs_names = self.adata.obs_names.values[:]

        x = np.array(self.adata.obs[time_key]).astype(self._dtype)

        adata = self.adata.raw.to_adata() if use_raw else self.adata
        gene_ix = np.where(adata.var_names == gene)[0]

        if data_key in ("X", None):
            y = adata.X[:, gene_ix]
            self._data_key = None
        elif data_key == "obs":
            y = adata.obs[gene].values
            # don't set data_key, it's just when `cell_color = ...`
        elif data_key in adata.layers:
            y = adata.layers[data_key][:, gene_ix]
            self._data_key = data_key
        else:
            raise NotImplementedError(f"Data key `{data_key!r}` is not implemented.")

        if lineage is not None:
            weight_threshold, val = weight_threshold
            w = _densify_squeeze(probs.X, self._dtype)
            w[w < weight_threshold] = val
        else:
            w = np.ones(len(x), dtype=self._dtype)

        if use_raw:
            correct_ixs = np.isin(self.adata.obs_names, adata.obs_names)
            x = x[correct_ixs]
            y = y[correct_ixs]
            w = w[correct_ixs]
            self._obs_names = self._obs_names[correct_ixs]

        del adata

        self._x_all, self._y_all, self._w_all = (
            _densify_squeeze(x, self._dtype)[:, np.newaxis],
            _densify_squeeze(y, self._dtype)[:, np.newaxis],
            w,
        )
        x, y, w = self.x_all[:], self.y_all[:], self.w_all[:]

        # sanity checks
        if self._x_all.shape[0] != self._y_all.shape[0]:
            raise ValueError(
                f"Independent variable's first dimension ({self._x_all.shape[0]}) "
                f"differs from dependent variable's first dimension ({self._y_all.shape[0]})."
            )
        if self._x_all.shape[0] != self._w_all.shape[0]:
            raise ValueError(
                f"Independent variable's first dimension ({self._x_all.shape[0]}) "
                f"differs from weights' first dimension ({self._w_all.shape[0]})."
            )

        x, ixs = np.unique(x, return_index=True)  # GAMR (mgcv) needs unique
        y = y[ixs]
        w = w[ixs]

        ixs = np.argsort(x)
        x, y, w = x[ixs], y[ixs], w[ixs]
        self._obs_names = self._obs_names[ixs]

        if np.allclose(w, w[0]):
            # degenerate case
            val_start = w[0] - 1
            val_end = w[0] + 1

        if val_start is None:
            val_start = np.min(x)
        if val_end is None:
            if threshold is None:
                threshold = np.nanmedian(w)
            # use `>=` because weights can all be 1
            w_test = w[w >= threshold]
            n_window = n_test_points // 20
            tmp = convolve(w_test, np.ones(n_window) / n_window, mode="nearest")
            val_end = x[w >= threshold][-1 if lineage is None else np.nanargmax(tmp)]

        if val_start > val_end:
            val_start, val_end = val_end, val_start
        val_start, val_end = (max(val_start, np.min(x)), min(val_end, np.max(x)))

        fil = (x >= val_start) & (x <= val_end)
        x_test = np.linspace(val_start, val_end, n_test_points) if n_test_points is not None else x[fil]
        x, y, w = x[fil], y[fil], w[fil]
        self._obs_names = self._obs_names[fil]

        if filter_cells is not None:
            tmp = y.squeeze()
            fil = (tmp >= filter_cells) & (~np.isclose(tmp, filter_cells).astype(bool))
            x, y, w = x[fil], y[fil], w[fil]
            self._obs_names = self._obs_names[fil]

        self._x, self._y, self._w = (
            self._reshape_and_retype(x),
            self._reshape_and_retype(y),
            self._reshape_and_retype(w).squeeze(-1),
        )
        self._x_test = self._reshape_and_retype(x_test)
        self._y_test = None
        self._x_hat = None
        self._y_hat = None
        self._conf_int = None

        if self.x.shape[0] == 0:
            raise RuntimeError("Unable to proceed, no values to fit.")

        if len({self.x.shape[0], self.y.shape[0], self.w.shape[0]}) != 1:
            raise RuntimeError(
                f"Values have different shapes: `{self.x.shape[0]}`, `{self.y.shape[0]}`, `{self.w.shape[0]}`."
            )

        self._gene = gene
        self._lineage = lineage
        self._prepared = True

        return self

    @abc.abstractmethod
    @d.get_sections(base="base_model_fit", sections=["Parameters"])
    @d.get_full_description(base="base_model_fit")
    def fit(
        self,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> "BaseModel":
        """Fit the model.

        Parameters
        ----------
        x
            Independent variables, array of shape ``(n_samples, 1)``. If :obj:`None`, use :attr:`x`.
        y
            Dependent variables, array of shape ``(n_samples, 1)``. If :obj:`None`, use :attr:`y`.
        w
            Optional weights of :attr:`x`, array of shape ``(n_samples,)``. If :obj:`None`, use :attr:`w`.
        kwargs
            Keyword arguments for underlying :attr:`model`'s fitting function.

        Returns
        -------
        Fits the :attr:`model` and returns self.
        """
        if not self.prepared:
            raise RuntimeError("The model has not been prepared yet, call `.prepare()` first.")

        self._check("_x", x)
        self._check("_y", y)
        self._check("_w", w, ndim=1)

        if self._x.shape != self._y.shape:
            raise ValueError(f"Inputs and targets differ in shape: `{self._x.shape}` vs. `{self._y.shape}`.")
        if self._y.shape[0] != self._w.shape[0]:
            raise ValueError(f"Inputs and weights differ in shape: `{self._y.shape[0]}` vs. `{self._w.shape[0]}`.")

        return self

    @abc.abstractmethod
    @d.get_sections(base="base_model_predict", sections=["Parameters", "Returns"])
    @d.get_full_description(base="base_model_predict")
    @d.dedent
    def predict(
        self,
        x_test: Optional[np.ndarray] = None,
        key_added: Optional[str] = "_x_test",
        **kwargs: Any,
    ) -> np.ndarray:
        """Run the prediction.

        Parameters
        ----------
        x_test
            Array of shape ``(n_samples,)`` used for prediction. If :obj:`None`, use :attr:`x_test`.
        key_added
            Attribute name where to save the :attr:`x_test` for later use. If :obj:`None`, don't save it.
        kwargs
            Keyword arguments for underlying :attr:`model`'s prediction method.

        Returns
        -------
        Returns and updates the following fields:

        - :attr:`y_test` - %(base_model_y_test.summary)s
        """

    @abc.abstractmethod
    @d.get_sections(base="base_model_ci", sections=["Parameters", "Returns"])
    @d.get_summary(base="base_model_ci")
    @d.get_full_description(base="base_model_ci")
    @d.dedent
    def confidence_interval(self, x_test: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """Calculate the confidence interval.

        Use :meth:`default_confidence_interval` function if underlying :attr:`model` has no method
        for confidence interval calculation.

        Parameters
        ----------
        x_test
            Array of shape ``(n_samples,)`` used for confidence interval calculation.
            If :obj:`None`, use :attr:`x_test`.
        kwargs
            Keyword arguments for underlying :attr:`model`'s confidence method
            or for :meth:`default_confidence_interval`.

        Returns
        -------
        Returns self and updates the following fields:

        - :attr:`conf_int` - %(base_model_conf_int.summary)s
        """

    @d.dedent
    def default_confidence_interval(
        self,
        x_test: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> np.ndarray:
        """Calculate the confidence interval, if the underlying :attr:`model` has no method for it.

        This formula is taken from :cite:`desalvo:70`, eq. 5.

        Parameters
        ----------
        %(base_model_ci.parameters)s

        Returns
        -------
        %(base_model_ci.returns)s

        Also updates the following fields:

        - :attr:`x_hat` - %(base_model_x_hat.summary)s
        - :attr:`y_hat` - %(base_model_y_hat.summary)s
        """
        use_ixs = self.w > 0
        x_hat = self.x[use_ixs]
        if x_test is None:
            x_test = self.x_test

        self._y_hat = self.predict(x_hat, key_added="_x_hat", **kwargs)
        self._y_test = self.predict(x_test, key_added="_x_test", **kwargs)

        n = np.sum(use_ixs)
        sigma_hat = np.sqrt(((self.y_hat - self.y[use_ixs].squeeze()) ** 2).sum() / (n - 2))
        mean = np.mean(self.x)

        # fmt: off
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            stds = sigma_hat * np.sqrt(1 + 1 / n + ((self.x_test - mean) ** 2) / ((self.x - mean) ** 2).sum())
        # fmt: on
        stds = np.squeeze(stds)

        self._conf_int = np.c_[self._y_test - stds / 2.0, self._y_test + stds / 2.0]

        return self.conf_int

    @d.dedent
    def plot(
        self,
        figsize: Tuple[float, float] = (8, 5),
        same_plot: bool = False,
        hide_cells: bool = False,
        perc: Tuple[float, float] = None,
        fate_prob_cmap: colors.ListedColormap = cm.viridis,
        cell_color: Optional[str] = None,
        lineage_color: str = "black",
        alpha: float = 0.8,
        lineage_alpha: float = 0.2,
        title: Optional[str] = None,
        size: int = 15,
        lw: float = 2,
        cbar: bool = True,
        margins: float = 0.015,
        xlabel: str = "pseudotime",
        ylabel: str = "expression",
        conf_int: bool = True,
        lineage_probability: bool = False,
        lineage_probability_conf_int: Union[bool, float] = False,
        lineage_probability_color: Optional[str] = None,
        obs_legend_loc: Optional[str] = "best",
        dpi: int = None,
        fig: mpl.figure.Figure = None,
        ax: mpl.axes.Axes = None,
        return_fig: bool = False,
        save: Optional[str] = None,
        **kwargs: Any,
    ) -> Optional[mpl.figure.Figure]:
        """Plot the smoothed gene expression.

        Parameters
        ----------
        figsize
            Size of the figure.
        same_plot
            Whether to plot all trends in the same plot.
        hide_cells
            Whether to hide the cells.
        perc
            Percentile by which to clip the fate probabilities.
        fate_prob_cmap
            Colormap to use when coloring in the fate probabilities.
        cell_color
            Key in :attr:`~anndata.AnnData.obs` or :attr:`~anndata.AnnData.var_names` used for coloring the cells.
        lineage_color
            Color for the lineage.
        alpha
            Alpha value in :math:`[0, 1]` for the transparency of cells.
        lineage_alpha
            Alpha value in :math:`[0, 1]` for the transparency lineage confidence intervals.
        title
            Title of the plot.
        size
            Size of the points.
        lw
            Line width for the smoothed values.
        cbar
            Whether to show the colorbar.
        margins
            Margins around the plot.
        xlabel
            Label on the x-axis.
        ylabel
            Label on the y-axis.
        conf_int
            Whether to show the confidence interval.
        lineage_probability
            Whether to show smoothed lineage probability as a dashed line.
            Note that this will require 1 additional model fit.
        lineage_probability_conf_int
            Whether to compute and show smoothed lineage probability confidence interval.
        lineage_probability_color
            Color to use when plotting the smoothed ``lineage_probability``.
            If :obj:`None`, it's the same as ``lineage_color``. Only used when ``show_lineage_probability = True``.
        obs_legend_loc
            Location of the legend when ``cell_color`` corresponds to a categorical variable.
        dpi
            Dots per inch.
        fig
            Figure to use. If :obj:`None`, create a new one.
        ax
            Ax to use. If :obj:`None`, create a new one.
        return_fig
            If :obj:`True`, return the figure object.
        save
            Filename where to save the plot. If :obj:`None`, just shows the plots.
        kwargs
            Keyword arguments for :meth:`~matplotlib.axes.Axes.legend`.

        Returns
        -------
        %(just_plots)s
        """
        if self.y_test is None:
            raise RuntimeError("Run `.predict()` first.")

        if fig is None or ax is None:
            fig, ax = plt.subplots(figsize=figsize, constrained_layout=True)

        if dpi is not None:
            fig.set_dpi(dpi)

        conf_int = conf_int and self.conf_int is not None
        hide_cells = hide_cells or self.x_all is None or self.w_all is None or self.y_all is None

        lineage_probability_color = lineage_color if lineage_probability_color is None else lineage_probability_color

        scaler = kwargs.pop(
            "scaler",
            self._create_scaler(
                lineage_probability,
                show_conf_int=conf_int,
            ),
        )

        if lineage_probability and ylabel in ("expression", self._gene):
            ylabel = f"scaled {ylabel}"

        vmin, vmax = None, None
        key, color, typp, mapper = self._get_colors(cell_color, same_plot=same_plot)

        if not hide_cells:
            cbar = cbar and (typp == ColorType.CONT)
            if typp == ColorType.CONT:
                vmin, vmax = _minmax(color, perc)

            _ = ax.scatter(
                self.x_all.squeeze(),
                scaler(self.y_all.squeeze()),
                c=color,
                s=size,
                cmap=fate_prob_cmap,
                vmin=vmin,
                vmax=vmax,
                alpha=alpha,
            )

        if title is None:
            title = f"{self._gene} @ {self._lineage}" if self._lineage is not None else f"{self._gene}"

        ax.plot(self.x_test, scaler(self.y_test), color=lineage_color, lw=lw, label=title)

        if title is not None:
            ax.set_title(title)
        if ylabel is not None:
            ax.set_ylabel(ylabel)
        if xlabel is not None:
            ax.set_xlabel(xlabel)

        ax.margins(margins)

        if conf_int:
            ax.fill_between(
                self.x_test.squeeze(),
                scaler(self.conf_int[:, 0]),
                scaler(self.conf_int[:, 1]),
                alpha=lineage_alpha,
                color=lineage_color,
                linestyle="--",
            )

        if lineage_probability and not isinstance(self, FittedModel) and not np.allclose(self.w, 1.0):
            from cellrank.pl._utils import _is_any_gam_mgcv

            model = copy.deepcopy(self)
            model._y = self._reshape_and_retype(self.w).copy()
            model = model.fit()

            if not lineage_probability_conf_int:
                y = model.predict()
            elif _is_any_gam_mgcv(model):
                y = model.predict(
                    level=lineage_probability_conf_int if isinstance(lineage_probability_conf_int, float) else 0.95
                )
            else:
                y = model.predict()
                model.confidence_interval()

                ax.fill_between(
                    model.x_test.squeeze(),
                    model.conf_int[:, 0],
                    model.conf_int[:, 1],
                    alpha=lineage_alpha,
                    color=lineage_probability_color,
                    linestyle="--",
                )

            handle = ax.plot(
                model.x_test,
                y,
                color=lineage_probability_color,
                lw=lw,
                linestyle="--",
                zorder=-1,
                label="probability",
            )

            self._maybe_add_legend(fig, ax, mapper=handle, title=None, **kwargs)

        if cbar and not hide_cells and not same_plot and not np.allclose(self.w_all, 1.0):
            norm = colors.Normalize(vmin=vmin, vmax=vmax)
            divider = make_axes_locatable(ax)
            cax = divider.append_axes("right", size="2%", pad=0.1)
            _ = mpl.colorbar.ColorbarBase(
                cax,
                norm=norm,
                cmap=fate_prob_cmap,
                ticks=np.linspace(norm.vmin, norm.vmax, 5),
            )
            cax.set_ylabel(key)
        elif typp == ColorType.CAT:
            self._maybe_add_legend(fig, ax, mapper, title=key, loc=obs_legend_loc, is_line=False)

        if save is not None:
            save_fig(fig, save)

        if return_fig:
            return fig

    def _maybe_add_legend(
        self,
        fig: mpl.figure.Figure,
        ax: mpl.axes.Axes,
        mapper: Union[Sequence[Any], Mapping[str, Any]],
        title: Optional[str] = None,
        loc: Optional[str] = "best",
        is_line: bool = True,
        **kwargs: Any,
    ) -> None:
        from cellrank.pl._utils import _position_legend

        if loc in ("none", None):
            return

        if isinstance(mapper, dict):
            handles = [
                ax.plot([], [], label=name, color=color)[0] if is_line else ax.scatter([], [], label=name, color=color)
                for name, color in mapper.items()
            ]
        else:
            handles = mapper

        legend = _position_legend(
            ax,
            legend_loc=loc,
            handles=handles,
            title=title,
            **kwargs,
        )
        fig.add_artist(legend)

    def _reshape_and_retype(self, arr: np.ndarray) -> np.ndarray:
        """Convert 1D or 2D array to 2D array of shape ``(n, 1)`` and set the data type.

        Parameters
        ----------
        arr
            Array to convert.

        Returns
        -------
        Array of shape ``(n, 1)`` with dtype set to :attr:`_dtype`.
        """
        if arr.ndim not in (1, 2):
            raise ValueError(f"Expected array to be 1 or 2 dimensional, found `{arr.ndim}` dimension(s).")
        if arr.ndim == 2 and arr.shape[1] != 1:
            raise ValueError(f"Expected the 2nd dimension to be 1, found `{arr.shape[1]}.`")

        return np.reshape(arr, (-1, 1)).astype(self._dtype)

    def _check(self, attr_name: Optional[str], arr: Optional[np.ndarray], ndim: int = 2) -> Optional[np.ndarray]:
        """Check if the attribute exists with the correct dimension and optionally set it.

        Parameters
        ----------
        attr_name
            Attribute name to check.
        arr
            Value to set. If :obj:`None`, just perform the checking.
        ndim
            Expected number of dimensions of the ``arr``.

        Returns
        -------
        The attribute under ``attr_name``.
        """
        if attr_name is None:
            return
        if arr is None:  # already called prepare
            if not hasattr(self, attr_name):
                raise AttributeError(f"No attribute `{attr_name!r}` found.")
            if not isinstance(getattr(self, attr_name, None), np.ndarray):
                raise AttributeError(
                    f"Expected `{attr_name!r}` to be `numpy.ndarray`, "
                    f"found `{type(getattr(self, attr_name, None)).__name__!r}`."
                )
            if getattr(self, attr_name).ndim != ndim:
                raise ValueError(
                    f"Expected attribute `{attr_name!r}` to have `{ndim}` dimensions, "
                    f"found `{getattr(self, attr_name).ndim}` dimensions."
                )
            return getattr(self, attr_name)

        setattr(self, attr_name, self._reshape_and_retype(arr))
        if attr_name.startswith("_"):
            try:
                getattr(self, attr_name[1:])
            except AttributeError:
                setattr(
                    self,
                    attr_name[1:],
                    property(lambda self: getattr(self, attr_name)),
                )
        return getattr(self, attr_name)

    def _deepcopy_attributes(self, dst: "BaseModel") -> None:
        # __deepcopy__ will usually call `_shallowcopy_attributes` twice, since it calls `.copy()`,
        # which should always call it (it copies the lineage and gene names + deepcopies the model)

        self._shallowcopy_attributes(dst)
        for attr in [
            "_x_all",
            "_y_all",
            "_w_all",
            "_x",
            "_y",
            "_w",
            "_x_test",
            "_y_test",
            "_x_hat",
            "_y_hat",
            "_conf_int",
            "_prepared",
        ]:
            setattr(dst, attr, copy.copy(getattr(self, attr)))

    def _shallowcopy_attributes(self, dst: "BaseModel") -> None:
        for attr in ["_gene", "_lineage", "_use_raw", "_data_key", "_is_bulk"]:
            setattr(dst, attr, copy.copy(getattr(self, attr)))
        # user is not exposed to this
        dst._obs_names = self._obs_names
        # always deepcopy the model (we're manipulating it in multiple threads/processed)
        dst._model = copy.deepcopy(self.model)

    @abc.abstractmethod
    @d.dedent
    def copy(self) -> "BaseModel":
        """%(copy)s"""  # noqa

    def __copy__(self) -> "BaseModel":
        return self.copy()

    def __deepcopy__(self, memodict={}) -> "BaseModel":  # noqa
        # deepcopy expects that `.copy()` makes a really shallow copy (i.e. only references to the arrays)
        # it should also not copy the `.prepared` attribute, since copying is happening mostly during
        # parallelization and it serves as 1 extra sanity check (a precaution that's not necessary, per-se, but highly
        # desirable)
        res = self.copy()
        # we don't copy the adata object for 2 reasons:
        # 1. these objects are meant to be as lightweight as possible
        # 2. in `.plot`, we deepcopy the model when plotting smoothed probabilities
        self._deepcopy_attributes(res)
        memodict[id(self)] = res

        return res

    def __bool__(self):
        return True

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return "<{}[gene={!r}, lineage={!r}, model={}]>".format(
            self.__class__.__name__,
            self._gene,
            self._lineage,
            None if self.model is None else _dup_spaces.sub(" ", str(self.model).replace("\n", " ")).strip(),
        )

    def _get_colors(
        self,
        key: Optional[str],
        *,
        same_plot: bool = False,
    ) -> Tuple[Optional[str], Optional[Union[str, np.ndarray]], ColorType, Optional[Dict[str, Any]],]:
        """
        Get color array.

        Parameters
        ----------
        key
            Key in :attr:`~anndata.AnnData.obs`, :attr:`~anndata.AnnData.var_names` or
            :attr:`~anndata.AnnData.raw.var_names`. Search first starts in `.obs`, then `.raw.var_names` and lastly
            `.layers`, using :meth:`~anndata.AnnData.obs_vector`.

        Returns
        -------
        Triple of the following:

        - name of the colorbar label.
        - array of values to map.
        - type of the color.
        - color mapper for categorical colors.
        """
        if colors.is_color_like(key):
            return None, key, ColorType.STR, None

        if key in self.adata.obs:
            if isinstance(self.adata.obs[key].dtype, pd.CategoricalDtype):
                add_colors_for_categorical_sample_annotation(
                    self.adata,
                    key=key,
                    force_update_colors=False,
                    palette=None,
                )
                col_dict = collections.defaultdict(
                    lambda: colors.to_rgb("grey"),
                    zip(
                        self.adata.obs[key].cat.categories,
                        [colors.to_rgb(i) for i in self.adata.uns[f"{key}_colors"]],
                    ),
                )
                return (
                    key,
                    np.array([col_dict[v] for v in self.adata.obs[key]]),
                    ColorType.CAT,
                    col_dict,
                )

            if np.issubdtype(self.adata.obs[key].dtype, np.number):
                return key, self.adata.obs[key].values, ColorType.CONT, None

            logg.debug(f"Unable to interpret cell color from type `{infer_dtype(self.adata.obs[key])}`")
            return None, "black", ColorType.STR, None

        if self._use_raw and key in self.adata.raw.var_names:
            return (
                key,
                _densify_squeeze(self.adata.raw[:, key], np.float64),
                ColorType.CONT,
                None,
            )

        try:
            # can in principle return data from `.obs`, in which case it's mostly invalid
            return (
                key,
                self.adata.obs_vector(key, layer=self._data_key),
                ColorType.CONT,
                None,
            )
        except KeyError:
            logg.debug(
                f"Key `{key!r}` not found in `adata.obs` or "
                f"`adata{'.raw' if self._use_raw else ''}.var_names`. Ignoring`"
            )

        if same_plot or np.allclose(self.w_all, 1.0):
            return None, "black", ColorType.STR, None

        return "fate probability", np.squeeze(self.w_all), ColorType.CONT, None

    def _create_scaler(self, show_lineage_probability: bool, show_conf_int: bool):
        if not show_lineage_probability:
            return lambda _: _

        minn, maxx = self._return_min_max(show_conf_int)
        return lambda x: (x - minn) / (maxx - minn)

    def _return_min_max(self, show_conf_int: bool):
        if self.y_test is None:
            raise RuntimeError("Run `.predict()` first.")

        vals = [self.y_test]
        # FittedModel Does not need to have these
        if self.y_all is not None:
            vals.append(self.y_all)
        if show_conf_int and self.conf_int is not None:
            vals.append(self.conf_int)

        minn = min(map(np.min, vals))
        maxx = max(map(np.max, vals))

        return minn, maxx


class FailedModel(BaseModel):
    """Model representing a failure of the original :attr:`model`.

    Parameters
    ----------
    model
        The original model which has failed.
    exc
        The exception that caused the :attr:`model` to fail or a :class:`str` containing the message.
        In the latter case, :meth:`~cellrank.models.FailedModel.reraise` a :class:`RuntimeError` with that message.
        If :obj:`None`, :class`UnknownModelError` will eventually be raised.
    """

    # in a functional programming language like Haskell essentially BaseModel would be a Maybe monad and this Nothing
    def __init__(self, model: BaseModel, exc: Optional[Union[BaseException, str]] = None):
        if not isinstance(model, BaseModel):
            raise TypeError(f"Expected `model` to be of type `BaseModel`, found `{type(model).__name__!r}`.")
        if exc is not None:
            if isinstance(exc, str):
                exc = RuntimeError(exc)
            if not isinstance(exc, BaseException):
                raise TypeError(
                    f"Expected `exc` to be either a string or a `BaseException`, found `{type(exc).__name__!r}`."
                )
        else:
            exc = UnknownModelError()

        super().__init__(model.adata, model)

        self._gene = model._gene
        self._lineage = model._lineage
        self._prepared = model._prepared
        self._is_bulk = model._is_bulk

        self._exc = exc

    def prepare(
        self,
        *_args,
        **_kwargs,
    ) -> "FailedModel":
        """Do nothing and return self."""
        return self

    def fit(
        self,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> "FailedModel":
        """Do nothing and return self."""
        return self

    def predict(
        self,
        x_test: Optional[np.ndarray] = None,
        key_added: Optional[str] = "_x_test",
        **kwargs: Any,
    ) -> np.ndarray:
        """Do nothing."""

    def confidence_interval(self, x_test: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """Do nothing."""

    def default_confidence_interval(
        self,
        x_test: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> np.ndarray:
        """Do nothing."""

    def reraise(self) -> None:
        """Raise the original exception with additional model information."""
        # retain the exception type and also the original exception
        raise type(self._exc)(f"Fatal model failure `{self}`.") from self._exc

    def _get_colors(
        self,
        key: Optional[str],
        *,
        same_plot: bool = False,
    ) -> Tuple[Optional[str], Optional[Union[str, np.ndarray]], ColorType, Optional[Dict[str, Any]],]:
        return None, "black", ColorType.STR, None

    def _return_min_max(self, show_conf_int: bool):
        return np.inf, -np.inf

    @d.dedent
    def copy(self) -> "FailedModel":
        """%(copy)s"""  # noqa
        return FailedModel(self.model.copy(), exc=self._exc)

    def __bool__(self):
        return False

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}[origin={repr(self.model).strip('<>')}]>"


@d.dedent
class FittedModel(BaseModel):
    """Class representing an already fitted model. Useful when the smoothed gene expression was computed externally.

    Parameters
    ----------
    x_test
        %(base_model_x_test.summary)s
    y_test
        %(base_model_y_test.summary)s
    conf_int
        %(base_model_conf_int.summary)s
        If :obj:`None`, always set ``conf_int=False`` in :meth:`~cellrank.models.BaseModel.plot`.
    x_all
        %(base_model_x_all.summary)s
        If :obj:`None`, always sets ``hide_cells=True`` in :meth:`~cellrank.models.BaseModel.plot`.
    y_all
        %(base_model_y_all.summary)s
        If :obj:`None`, always sets `hide_cells=True`` in :meth:`~cellrank.models.BaseModel.plot`.
    w_all
        %(base_model_w_all.summary)s
        If :obj:`None` and :attr:`x_all` and :attr:`y_all` are present, it will be set an array of :math:`1`.
    """

    def __init__(
        self,
        x_test: ArrayLike,
        y_test: ArrayLike,
        conf_int: Optional[ArrayLike] = None,
        x_all: Optional[ArrayLike] = None,
        y_all: Optional[ArrayLike] = None,
        w_all: Optional[ArrayLike] = None,
    ):
        super().__init__(None, None)

        self._x_test = _densify_squeeze(x_test, self._dtype)[:, np.newaxis]
        self._y_test = _densify_squeeze(y_test, self._dtype)

        if self.x_test.ndim != 2 or self.x_test.shape[1] != 1:
            raise ValueError(f"Expected `x_test` to be of shape `(..., 1)`, found `{self.x_test.shape}`.")

        if self.y_test.shape != (self.x_test.shape[0],):
            raise ValueError(
                f"Expected `y_test` to be of shape `({self.x_test.shape[0]},)`, " f"found `{self.y_test.shape}`."
            )

        if conf_int is not None:
            self._conf_int = _densify_squeeze(conf_int, self._dtype)
            if self.conf_int.shape != (self.x_test.shape[0], 2):
                raise ValueError(f"Expected `conf_int` to be of shape `({self.x_test.shape[0]}, 2)`.")
        else:
            logg.debug("No `conf_int` have been supplied, will be ignored during plotting")

        if x_all is not None and y_all is not None:
            self._x_all = _densify_squeeze(x_all, self._dtype)[:, np.newaxis]
            if self.x_all.ndim != 2 or self.x_all.shape[1] != 1:
                raise ValueError(f"Expected `x_all` to be of shape `(..., 1)`, found `{self.x_all.shape}`.")

            self._y_all = _densify_squeeze(y_all, self._dtype)[:, np.newaxis]
            if self.y_all.shape != self.x_all.shape:
                raise ValueError(
                    f"Expected `y_all` to be of shape `({self.x_all.shape[0]}, 1)`, found `{self.y_all.shape}`."
                )

            if w_all is not None:
                self._w_all = _densify_squeeze(w_all, self._dtype)
                if self.w_all.ndim != 1 or self.w_all.shape[0] != self.x_all.shape[0]:
                    raise ValueError(
                        f"Expected `w_all` to be of shape `({self.x_all.shape[0]},)`, " f"found `{self.w_all.shape}`."
                    )
            else:
                logg.debug("Setting `w_all` to an array of `1`")
                self._w_all = np.ones_like(self.x_all, dtype=np.float64).squeeze(1)
        else:
            logg.debug(
                "None or partially incomplete `x_all` and `y_all` have been supplied, "
                "will be ignored during plotting"
            )

        self._prepared = True

    def _get_colors(
        self,
        key: Optional[str],
        *,
        same_plot: bool = False,
    ) -> Tuple[Optional[str], Optional[Union[str, np.ndarray]], ColorType, Optional[Dict[str, Any]],]:
        # w_all does not need to be defined
        if same_plot or self.w_all is None or np.allclose(self.w_all, 1.0):
            return None, "black", ColorType.STR, None
        return "fate probability", np.squeeze(self.w_all), ColorType.CONT, None

    def prepare(self, *_args, **_kwargs) -> "FittedModel":
        """Do nothing and return self."""
        return self

    def fit(
        self,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> "FittedModel":
        """Do nothing and return self."""
        return self

    @d.dedent
    def predict(
        self,
        x_test: Optional[np.ndarray] = None,
        key_added: Optional[str] = "_x_test",
        **kwargs: Any,
    ) -> np.ndarray:
        """%(base_model_y_test.summary)s."""
        return self._y_test

    @d.dedent
    def confidence_interval(self, x_test: Optional[np.ndarray] = None, **kwargs) -> np.ndarray:
        """%(base_model_conf_int.summary)s Raise a :class:`RuntimeError` if not present."""
        if self.conf_int is None:
            raise RuntimeError(
                "No confidence interval has been supplied. " "Use `conf_int=...` when instantiating this class."
            )
        return self.conf_int

    @d.dedent
    def default_confidence_interval(
        self,
        x_test: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> np.ndarray:
        """%(base_model_conf_int.summary)s Raise a :class:`RuntimeError` if not present."""
        return self.confidence_interval()

    @d.dedent
    def copy(self) -> "FittedModel":
        """%(copy)s"""  # noqa
        # here we return a deepcopy since it doesn't make sense to make a shallow one
        return FittedModel.from_model(self)

    @staticmethod
    def from_model(model: BaseModel) -> "FittedModel":
        """Create a :class:`~cellrank.models.FittedModel` from a :class:`~cellrank.models.BaseModel`."""
        if not isinstance(model, BaseModel):
            raise TypeError(f"Expected `model` to be of type `BaseModel`, found `{type(model).__name__!r}`.")

        fm = FittedModel(
            model.x_test,
            model.y_test,
            conf_int=model.conf_int,
            x_all=model.x_all,
            y_all=model.y_all,
            w_all=model.w_all,
        )

        fm._gene = model._gene
        fm._lineage = model._lineage
        fm._is_bulk = model._is_bulk
        fm._prepared = True

        return fm



import abc
import contextlib
import copy as copy_
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
import pandas as pd
import scipy.sparse as sp

from anndata import AnnData

from cellrank._utils._docs import d
from cellrank._utils._key import Key
from cellrank._utils._lineage import Lineage
from cellrank.estimators.mixins import KernelMixin
from cellrank.kernels import PrecomputedKernel
from cellrank.kernels._base_kernel import KernelExpression
from cellrank.kernels.mixins import AnnDataMixin, IOMixin

__all__ = ["BaseEstimator"]

Attr_t = (Literal["X", "raw", "layers", "obs", "var", "obsm", "varm", "obsp", "varp", "uns"],)


@d.get_sections(base="base_estimator", sections=["Parameters"])
class BaseEstimator(IOMixin, KernelMixin, AnnDataMixin, abc.ABC):
    """Base class for all estimators.

    Parameters
    ----------
    object
        Can be one of the following types:

        - :class:`~anndata.AnnData` - annotated data object.
        - :class:`~scipy.sparse.spmatrix`, :class:`~numpy.ndarray` - row-normalized transition matrix.
        - :class:`~cellrank.kernels.KernelExpression` - kernel expression.
        - :class:`str` - key in :attr:`~anndata.AnnData.obsp` where the transition matrix is stored and
          ``adata`` must be provided in this case.
        - :class:`bool` - directionality of the transition matrix that will be used to infer its storage location.
          If :obj:`None`, the directionality will be determined automatically and
          ``adata`` must be provided in this case.
    kwargs
        Keyword arguments for the :class:`~cellrank.kernels.PrecomputedKernel`.
    """

    def __init__(
        self,
        object: Union[str, bool, np.ndarray, sp.spmatrix, AnnData, KernelExpression],
        **kwargs: Any,
    ):
        if isinstance(object, KernelExpression):
            if object.transition_matrix is None:
                raise RuntimeError("Compute transition matrix first as `.compute_transition_matrix()`.")
        else:
            object = PrecomputedKernel(object, copy=False, **kwargs)
        super().__init__(kernel=object)

        self._params: Dict[str, Any] = {}
        self._shadow_adata = AnnData(
            X=sp.csr_matrix(self.adata.shape, dtype=self.adata.X.dtype),
            obs=self.adata.obs[[]].copy(),
            var=self.adata.var[[]].copy(),
            raw=None if self.adata.raw is None else self.adata.raw.to_adata(),
        )

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__()

    @abc.abstractmethod
    def fit(self, *args: Any, **kwargs: Any) -> "BaseEstimator":
        """Fit the estimator.

        Parameters
        ----------
        args
            Positional arguments.
        kwargs
            Keyword arguments.

        Returns
        -------
        Self.
        """

    @abc.abstractmethod
    def predict(self, *args: Any, **kwargs: Any) -> "BaseEstimator":
        """Run the prediction.

        Parameters
        ----------
        args
            Positional arguments.
        kwargs
            Keyword arguments.

        Returns
        -------
        Self.
        """

    def _set(
        self,
        attr: Optional[str] = None,
        obj: Optional[Union[pd.DataFrame, Mapping[str, Any]]] = None,
        key: Optional[str] = None,
        value: Optional[Union[np.ndarray, pd.Series, pd.DataFrame, Lineage, AnnData, Dict[str, Any]]] = None,
        copy: bool = True,
        shadow_only: bool = False,
    ) -> None:
        """Set an attribute and optionally update ``obj['{key}']``.

        Parameters
        ----------
        attr
            Attribute to set. Only updated when we're not in the shadow. If :obj:`None`, don't update anything.
            See :attr:`_in_shadow` and ``obj`` for more information.
        obj
            Object which to update with ``value`` alongside the ``attr``.
            Usually, an attribute of :attr:`adata` is passed here.
        key
            Key in ``obj`` to update with ``value``. Only used when ``obj != None``.
        value
            Value to set. If :obj:`None` and ``key != None``, it removes the values under ``obj['{key}']``, if present.
        copy
            Whether to copy the ``value`` before setting it in ``obj``.
        shadow_only
            Whether to update the ``obj`` if we are not in the shadow.

        Returns
        -------
        Nothing, just optionally updates ``attr`` and/or ``obj[{key}]``.

        Raises
        ------
        AttributeError
            If ``attr`` doesn't exist.
        """
        if not self._in_shadow:
            if attr is not None:
                if not hasattr(self, attr):
                    raise AttributeError(attr)
                setattr(self, attr, value)
            if shadow_only:
                return

        if obj is None:
            return

        if key is not None:
            if value is None:
                with contextlib.suppress(KeyError):
                    del obj[key]

            else:
                obj[key] = copy_.copy(value) if copy else value

    def _get(
        self,
        *,
        obj: Union[pd.DataFrame, Mapping[str, Any]],
        key: str,
        shadow_attr: Optional[Literal["obs", "obsm", "var", "varm", "uns"]] = None,
        dtype: Optional[Union[type, Tuple[type, ...]]] = None,
        copy: bool = True,
        allow_missing: bool = False,
    ) -> Any:
        """Get data from an object and set an attribute.

        Parameters
        ----------
        obj
            Object from which to extract the data.
        key
            Key in ``obj`` where the data is stored.
        shadow_attr
            Attribute of :attr:`_shadow_adata` where to save the extracted data. If :obj:`None`, don't update it.
        dtype
            Valid type(s) of the extracted data.
        copy
            Copy the data before setting the ``self_attr``.
        allow_missing
            Whether to allow ``key`` to be missing in ``obj``.

        Returns
        -------
        The extracted values and optionally updates :attr:`_shadow_adata` ``.{shadow_attr}``.

        Raises
        ------
        AttributeError
            If ``attr`` doesn't exist.
        TypeError
            If ``dtype != None`` and the extracted values are not instances of ``dtype``.
        KeyError
            If ``allow_missing = False`` and ``key`` was not found in ``obj``.
        """
        if shadow_attr is not None and not hasattr(self._shadow_adata, shadow_attr):
            raise AttributeError(shadow_attr)

        try:
            data = obj[key]
            if dtype is not None and not isinstance(data, dtype):
                raise TypeError(f"Expected object to be of type `{dtype}`, found `{type(data).__name__}`.")
            if copy:
                data = copy_.copy(data)
            if shadow_attr is not None:
                getattr(self._shadow_adata, shadow_attr)[key] = data
            return data
        except KeyError:
            if not allow_missing:
                raise
            return None

    @property
    @contextlib.contextmanager
    def _shadow(self) -> None:
        """Temporarily set :attr:`adata` to :attr:`_shadow_adata`.

        Used to construct the serialization object in :meth:`to_adata`.
        """
        if self._in_shadow:
            yield
        else:
            adata = self.adata
            try:
                self.adata = self._shadow_adata
                yield
            finally:
                self.adata = adata

    @property
    def _in_shadow(self) -> bool:
        """Return `True` if :attr:`adata` is :attr:`_shadow_adata`."""
        return self.adata is self._shadow_adata

    def _create_params(
        self,
        locs: Optional[Mapping[str, Any]] = None,
        func: Optional[Callable] = None,
        remove: Sequence[str] = (),
    ) -> Dict[str, Any]:
        """Create parameters of interest from a function call.

        Parameters
        ----------
        locs
            Environment from which to get the parameters. If `None`, get the caller's environment.
        func
            Function of interest. If :obj:`None`, use the caller.
        remove
            Keys in ``locs`` which should not be included in the result.

        Returns
        -------
        The parameters as a :class:`dict`.

        Notes
        -----
        *args/**kwargs are always ignored and the values in ``locs`` are not copied.
        """
        frame = inspect.currentframe()
        try:
            if locs is None:
                locs = frame.f_back.f_locals
            if func is None:
                name = frame.f_back.f_code.co_name
                func = dict(inspect.getmembers(self)).get(name, None)
            if not callable(func):
                raise TypeError(f"Expected `func` to be `callable`, found `{type(func).__name__}`.")

            params = {}
            for name, param in inspect.signature(func).parameters.items():
                if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
                    continue
                if name in remove:
                    continue
                if name in locs:
                    params[name] = locs[name]

            return params
        except AttributeError:
            # frame can be None
            return {}
        except TypeError:
            return {}
        finally:
            del frame

    def _read_params(self, key: str) -> Dict[str, Any]:
        """Read ``key`` from estimator params in :attr:`adata`.

        Usually called in :meth:`_read_adata` during :meth:`from_adata`.
        """
        ekey = Key.uns.estimator(self.backward) + "_params"
        return dict(self.adata.uns.get(ekey, {}).get(key, {}))

    @d.dedent
    def to_adata(
        self,
        keep: Union[Literal["all"], Sequence[Attr_t]] = ("X", "raw"),
        *,
        copy: Union[bool, Sequence[Attr_t]] = True,
    ) -> AnnData:
        """%(to_adata.full_desc)s

        Parameters
        ----------
        keep
            Which attributes to keep from the underlying :attr:`adata`. Valid options are:

            - ``'all'`` - keep all attributes specified in the signature.
            - :class:`~typing.Sequence` - keep only subset of these attributes.
            - :class:`dict` - the keys correspond the attribute names and values to a subset of keys
              which to keep from this attribute. If the values are specified either as :obj:`True` or ``'all'``,
              everything from this attribute will be kept.
        copy
            Whether to copy the data. Can be specified on per-attribute basis.
            Useful for attributes that are array-like.

        Returns
        -------
        Annotated data object.
        """  # noqa: D400

        def handle_attribute(attr: Attr_t, keys: List[str], *, copy: bool) -> None:
            try:
                if attr == "X":
                    adata.X = copy_.deepcopy(self.adata.X) if copy else self.adata.X
                    return
                if attr == "raw":
                    adata.raw = self.adata.raw.to_adata() if self.adata.raw is not None else None
                    return

                old = getattr(self.adata, attr)
                new = getattr(adata, attr)
                if keys == ["all"]:
                    keys = list(old.keys())

                # fmt: off
                if isinstance(new, pd.DataFrame):
                    old = old[keys]
                    # avoid duplicates
                    old = old[old.columns.difference(new.columns)]
                    setattr(adata, attr, pd.merge(new, old, how="inner", left_index=True, right_index=True, copy=copy))
                elif isinstance(new, Mapping):
                    old = {k: old[k] for k in keys}
                    # old has preference, since it's user supplied
                    setattr(adata, attr, {**new, **(copy_.deepcopy(old) if copy else old)})
                else:
                    raise TypeError(f"Expected `adata.{attr}` to be either `Mapping` or `pandas. DataFrame`, "
                                    f"found `{type(new).__name__}`.")
                # fmt: on
            except KeyError:
                missing = sorted(k for k in keys if k not in old)
                raise KeyError(f"Unable to find key(s) `{missing}` in `adata.{attr}`.") from None

        adata = self._shadow_adata.copy()
        _adata = self.adata
        try:
            # kernel and estimator share the adata
            self.adata = adata
            self.kernel.write_to_adata()
        finally:
            self.adata = _adata
        key = Key.uns.estimator(self.backward) + "_params"
        adata.uns[key] = copy_.deepcopy(self.params)

        # fmt: off
        if isinstance(keep, str):
            if keep == "all":
                keep = ["X", "raw", "layers", "obs", "var", "obsm", "varm", "obsp", "varp", "uns"]
            else:
                keep = [keep]
        if not isinstance(keep, Mapping):
            keep = {attr: True for attr in keep}
        if isinstance(copy, bool):
            copy = {attr: copy for attr in keep}
        elif isinstance(copy, str):
            copy = [copy]
        if not isinstance(copy, Mapping):
            copy = {attr: True for attr in copy}
        # fmt: on

        for attr, keys in keep.items():
            if keys is True:
                keys = ["all"]
            elif isinstance(keys, str):
                keys = [keys]
            if keys is False or not len(keys):
                continue
            handle_attribute(attr, keys=list(keys), copy=copy.get(attr, False))

        return adata

    @classmethod
    @d.dedent
    def from_adata(cls, adata: AnnData, obsp_key: str) -> "BaseEstimator":
        """%(from_adata.full_desc)s

        Parameters
        ----------
        %(adata)s
        obsp_key
            Key in :attr:`~anndata.AnnData.obsp` where the transition matrix is stored.

        Returns
        -------
        %(from_adata.returns)s
        """  # noqa: D400
        return super().from_adata(adata, obsp_key=obsp_key)

    @d.dedent
    def copy(self, *, deep: bool = False) -> "BaseEstimator":
        """Return a copy of self.

        Parameters
        ----------
        deep
            Whether to return a deep copy or not. If :obj:`True`, this also copies the :attr:`adata`.

        Returns
        -------
        A copy of self.
        """
        k = copy_.deepcopy(self.kernel) if deep else copy_.copy(self.kernel)
        res = type(self)(k)
        for k, v in self.__dict__.items():
            if isinstance(v, Mapping):
                res.__dict__[k] = copy_.deepcopy(v)
            elif k != "_kernel":
                res.__dict__[k] = copy_.deepcopy(v) if deep else copy_.copy(v)

        return res

    def __copy__(self) -> "BaseEstimator":
        return self.copy(deep=False)

    def _format_params(self) -> str:
        return f"kernel={self.kernel!s}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}[{self._format_params()}]"

    def __str__(self) -> str:
        return repr(self)

    @property
    def params(self) -> Dict[str, Any]:
        """Estimator parameters."""
        return self._params



import abc
from typing import Any, Optional

import numpy as np
import pandas as pd

from matplotlib.colors import Normalize, to_hex
from matplotlib.pyplot import get_cmap

from anndata import AnnData

from cellrank._utils._docs import d
from cellrank.kernels._base_kernel import BidirectionalKernel
from cellrank.kernels._utils import _ensure_numeric_ordered, require_tmat

__all__ = ["ExperimentalTimeKernel"]


@d.dedent
class ExperimentalTimeKernel(BidirectionalKernel, abc.ABC):
    """Kernel base class which computes transition probabilities based on experimental time.

    %(density_correction)s

    Parameters
    ----------
    %(adata)s
    %(backward)s
    time_key
        Key in :attr:`~anndata.AnnData.obs` where experimental time is stored.
        The experimental time can be of either of a numeric or an ordered categorical type.
    kwargs
        Keyword arguments for the parent class.
    """

    def __init__(
        self,
        adata: AnnData,
        time_key: str,
        backward: bool = False,
        **kwargs: Any,
    ):
        super().__init__(
            adata,
            backward=backward,
            time_key=time_key,
            **kwargs,
        )

    def _read_from_adata(self, time_key: str, cmap: str = "gnuplot", **kwargs: Any) -> None:
        super()._read_from_adata(**kwargs)

        self._time_key = time_key
        self._exp_time = _ensure_numeric_ordered(self.adata, time_key)
        self.adata.obs[time_key] = self.experimental_time.values

        # fmt: off
        cmap = get_cmap(cmap)
        cats = self.experimental_time.cat.categories
        norm = Normalize(vmin=cats.min(), vmax=cats.max())
        self.adata.uns[f"{time_key}_colors"] = np.array([to_hex(c) for c in cmap(norm(cats))])
        # fmt: on

    @d.dedent
    @require_tmat
    def plot_single_flow(
        self,
        cluster: str,
        cluster_key: str,
        time_key: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """%(plot_single_flow.full_desc)s

        Parameters
        ----------
        %(plot_single_flow.parameters)s

        Returns
        -------
        %(plot_single_flow.returns)s
        """  # noqa: D400
        if time_key is None:
            time_key = self._time_key
        return super().plot_single_flow(cluster, cluster_key, time_key, *args, **kwargs)

    @property
    def experimental_time(self) -> pd.Series:
        """Experimental time."""
        return self._exp_time

    def __invert__(self, *args: str) -> "ExperimentalTimeKernel":
        etk = self._copy_ignore(*(("_transition_matrix",) + args))
        minn, maxx = etk.experimental_time.min(), etk.experimental_time.max()
        etk._exp_time = pd.Series(
            pd.Categorical(maxx - np.array(etk.experimental_time) + minn, ordered=True),
            index=etk.experimental_time.index,
        )
        etk._backward = not self.backward
        etk._params = {}
        return etk



import inspect
import warnings
from typing import Any, Iterable, Optional

import numpy as np
from sklearn.base import BaseEstimator
from sklearn.utils import check_X_y

from anndata import AnnData

from cellrank._utils._docs import d
from cellrank.models import BaseModel

__all__ = ["SKLearnModel"]


@d.dedent
class SKLearnModel(BaseModel):
    """Wrapper around :class:`~sklearn.base.BaseEstimator`.

    Parameters
    ----------
    %(adata)s
    model
        Instance of the underlying :mod:`sklearn` estimator, such as :class:`~sklearn.svm.SVR`.
    weight_name
        Name of the weight argument when fitting the model. If :obj:`None`, to determine it automatically.
        If an empty :class:`str`, no weights will be used.
    ignore_raise
        Do not raise an exception if weight argument is not found when fitting the :attr:`model`.
        This is useful in case when the weight argument is passed in the ``**kwargs`` and
        cannot be determined from signature.
    """

    _fit_names = ("fit", "__init__")
    _predict_names = ("predict", "__call__")
    _weight_names = ("w", "weights", "sample_weight", "sample_weights")
    _conf_int_names = ("conf_int", "confidence_intervals")

    def __init__(
        self,
        adata: AnnData,
        model: BaseEstimator,
        weight_name: Optional[str] = None,
        ignore_raise: bool = False,
    ):
        if not isinstance(model, BaseEstimator):
            raise TypeError(f"Expected model to be of type `BaseEstimator`, found `{type(model).__name__!r}`.")

        super().__init__(adata, model)

        fit_name = self._find_func(self._fit_names)
        predict_name = self._find_func(self._predict_names)
        ci_name = self._find_func(self._conf_int_names, use_default=True, default=None)

        self._weight_name = self._find_arg_name(fit_name, self._weight_names) if weight_name is None else weight_name

        if self._weight_name is None:
            raise RuntimeError(
                f"Unable to determine weights for function `{fit_name!r}`, searched `{self._weight_names}`. "
                f"Consider specifying it manually as `weight_name=...`."
            )
        if (
            not ignore_raise
            and self._weight_name != ""
            and self._weight_name not in inspect.signature(getattr(self.model, fit_name)).parameters
        ):
            raise ValueError(
                f"Unable to detect `{weight_name!r}` in the signature of `{fit_name!r}`."
                f"If it's in `kwargs`, set `ignore_raise=True.`"
            )

        self._fit_fn = getattr(self.model, fit_name)
        self._pred_fn = getattr(self.model, predict_name)
        self._ci_fn = None if ci_name is None else getattr(self.model, ci_name, None)

    @d.dedent
    def fit(
        self,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> "SKLearnModel":
        """%(base_model_fit.full_desc)s

        Parameters
        ----------
        %(base_model_fit.parameters)s

        Returns
        -------
        Fits the model and returns self.
        """  # noqa: D400
        super().fit(x, y, w, **kwargs)

        if self._weight_name not in (None, ""):
            kwargs[self._weight_name] = self._w

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            x, y = check_X_y(
                X=self.x,
                y=self.y,
                force_all_finite="allow-nan",
                copy=False,
                estimator=self.model,
            )
        self._model = self._fit_fn(x, y, **kwargs)

        return self

    @d.dedent
    def predict(self, x_test: Optional[np.ndarray] = None, key_added: str = "_x_test", **kwargs) -> np.ndarray:
        """%(base_model_predict.full_desc)s

        Parameters
        ----------
        %(base_model_predict.parameters)s

        Returns
        -------
        %(base_model_predict.returns)s
        """  # noqa: D400
        x_test = self._check(key_added, x_test)

        self._y_test = self._pred_fn(x_test, **kwargs)
        self._y_test = np.squeeze(self._y_test).astype(self._dtype)

        return self.y_test

    @d.dedent
    def confidence_interval(self, x_test: Optional[np.ndarray] = None, **kwargs: Any) -> np.ndarray:
        """%(base_model_ci.full_desc)s

        Parameters
        ----------
        %(base_model_ci.parameters)s

        Returns
        -------
        %(base_model_ci.returns)s
        """  # noqa: D400
        if self._ci_fn is None:
            return self.default_confidence_interval(x_test=x_test, **kwargs)

        x_test = self._check("_x_test", x_test)
        self._conf_int = self._ci_fn(x_test, **kwargs)

        return self.conf_int

    def _find_func(
        self,
        func_names: Iterable[str],
        use_default: bool = False,
        default: Optional[str] = None,
    ) -> Optional[str]:
        """Find a function in :attr:`model` from given names.

        If :obj:`None` is found, use the ``default`` or raise a :class:`RuntimeError`.

        Parameters
        ----------
        func_names
            Function names to search. The first one found is returned.
        use_default
            Whether to return the ``default`` if it is :obj:`None` or raise :class:`RuntimeError`.
        default
            The default function name to use if :obj:`None` was found.

        Returns
        -------
        Name of the function or the default name.
        """
        for name in func_names:
            if hasattr(self.model, name) and callable(getattr(self.model, name)):
                return name
        if use_default:
            return default
        raise RuntimeError(f"Unable to find function and no default specified, searched for `{list(func_names)}`.")

    def _find_arg_name(self, func_name: Optional[str], param_names: Iterable[str]) -> Optional[str]:
        """Find an argument in :attr:`model`'s ``func_name``.

        Parameters
        ----------
        func_name
            Function name of :attr:`model`.
        param_names
            Parameter names to search. The first one found is returned.

        Returns
        -------
        The parameter name or :obj:`None`, if :obj:`None` was found or ``func_name`` was :obj:`None`.
        """
        if func_name is None:
            return None

        for param in inspect.signature(getattr(self.model, func_name)).parameters:
            if param in param_names:
                return param

        return None

    @property
    def model(self) -> BaseEstimator:
        """The underlying :class:`~sklearn.base.BaseEstimator`."""
        return self._model

    @d.dedent
    def copy(self) -> "SKLearnModel":
        """%(copy)s"""  # noqa
        res = SKLearnModel(self.adata, self._model, weight_name=self._weight_name, ignore_raise=True)
        self._shallowcopy_attributes(res)  # this deepcopies the underlying model

        return res



import copy
import enum
from typing import Any, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd
import scipy.stats as st

from anndata import AnnData

from cellrank._utils._docs import d, inject_docs
from cellrank._utils._enum import ModeEnum
from cellrank.models import BaseModel, FailedModel
from cellrank.models._utils import _OFFSET_KEY, _get_knotlocs, _get_offset

__all__ = ["GAMR"]

# cache so we don't have to re-load
_r_lib = None
_r_lib_name = None


class KnotLocs(ModeEnum):
    AUTO = enum.auto()
    DENSITY = enum.auto()


@inject_docs(key=_OFFSET_KEY, kloc=KnotLocs)
@d.dedent
class GAMR(BaseModel):
    """Wrapper around R's `mgcv <https://cran.r-project.org/web/packages/mgcv/>`__ package for fitting GAMs.

    Parameters
    ----------
    %(adata)s
    n_knots
        Number of knots.
    distribution
        Distribution family in `rpy2.robjects.r`, such as ``'gaussian'`` or ``'nb'`` for negative binomial.
        If ``'nb'``, raw count data in :attr:`~anndata.AnnData.raw` is always used.
    basis
        Basis for the smoothing term.
        See `here <https://www.rdocumentation.org/packages/mgcv/versions/1.8-33/topics/s>`__ for valid options.
    knotlocs
        Position of the knots. Can be one of the following:

        - ``{kloc.AUTO!r}`` - let `mgcv` handle the knot positions.
        - ``{kloc.DENSITY!r}`` - position the knots based on the density of the pseudotime.
    offset
        Offset term for the GAM. Only available when ``distribution='nb'``. If `'default'`, it is calculated
        according to :cite:`robinson:10`. The values are saved in :attr:`adata.obs['{key}'] <anndata.AnnData.obs>`.
        If :obj:`None`, no offset is used.
    smoothing_penalty
        Penalty for the smoothing term. The larger the value, the smoother the fitted curve.
    kwargs
        Keyword arguments for `gam control
        <https://www.rdocumentation.org/packages/mgcv/versions/1.8-33/topics/gam.control>`__.
    """

    def __init__(
        self,
        adata: AnnData,
        n_knots: int = 5,
        distribution: str = "gaussian",
        basis: str = "cr",
        knotlocs: Literal["auto", "density"] = KnotLocs.AUTO,
        offset: Optional[Union[np.ndarray, Literal["default"]]] = "default",
        smoothing_penalty: float = 1.0,
        **kwargs: Any,
    ):
        if n_knots <= 0:
            raise ValueError(f"Expected `n_splines` to be positive, found `{n_knots}`.")
        if smoothing_penalty < 0:
            raise ValueError(f"Expected `smoothing_penalty` to be non-negative, found `{smoothing_penalty}`.")

        super().__init__(adata, model=None)
        self._n_knots = n_knots
        self._family = distribution

        self._formula = f"y ~ s(x, bs={basis!r}, k={self._n_knots}, sp={smoothing_penalty})"
        self._design_mat = None
        self._offset = None
        self._knotslocs = KnotLocs(knotlocs)

        self._control_kwargs = copy.copy(kwargs)

        self._lib = None
        self._lib_name = None

        # it's a bit costly to import, copying just passes the reference
        if kwargs.pop("perform_import_check", True):
            self._lib, self._lib_name = _maybe_import_r_lib("mgcv")

        if distribution == "nb" and offset is not None:
            if not isinstance(offset, (np.ndarray, str)):
                raise TypeError(
                    f"Expected `offset` to be either `'default'` or `numpy.ndarray`," f"got `{type(offset).__name__}`."
                )

            if isinstance(offset, str):
                if offset != "default":
                    raise ValueError("Only value `'default'` is allowed when `offset` is a string.")
                offset = _get_offset(adata, use_raw=True, recompute=False)

            offset = np.asarray(offset, dtype=self._dtype)

            if offset.shape != (adata.n_obs,):
                raise ValueError(f"Expected offset to be of shape `{(adata.n_obs,)}`, found `{offset.shape}`.")
            adata.obs[_OFFSET_KEY] = offset

            self._formula += " + offset(offset)"
            self._offset = offset

    @d.dedent
    def prepare(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> "GAMR":
        """%(base_model_prepare.full_desc)s

        This also removes the zero and negative weights and prepares the design matrix.

        Parameters
        ----------
        %(base_model_prepare.parameters)s

        Returns
        -------
        %(base_model_prepare.returns)s
        """  # noqa

        if self._family == "nb":
            kwargs["use_raw"] = True
        prepared = super().prepare(*args, **kwargs)
        if isinstance(prepared, FailedModel):
            return prepared

        use_ixs = self.w > 0
        self._x = self.x[use_ixs]
        self._y = self.y[use_ixs]
        self._w = self.w[use_ixs]

        self._design_mat = pd.DataFrame(
            np.c_[self.x, self.y],
            columns=["x", "y"],
        )

        if self._offset is not None:
            mask = np.isin(self.adata.obs_names, self._obs_names[use_ixs])
            self._design_mat["offset"] = self._offset[mask]

        return self

    @d.dedent
    def fit(
        self,
        x: Optional[np.ndarray] = None,
        y: Optional[np.ndarray] = None,
        w: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> "GAMR":
        """%(base_model_fit.full_desc)s

        Parameters
        ----------
        %(base_model_fit.parameters)s

        Returns
        -------
        Fits the model and returns self. Updates the following fields by filtering out :math:`0` weights :attr:`w`:

        - :attr:`x` - %(base_model_x.summary)s
        - :attr:`y` - %(base_model_y.summary)s
        - :attr:`w` - %(base_model_w.summary)s
        """  # noqa

        import rpy2.robjects as ro
        from rpy2.robjects import pandas2ri
        from rpy2.robjects.conversion import localconverter

        super().fit(x, y, w, **kwargs)

        # order seems to matter, pandas2ri + default_converter results in:
        # Conversion 'py2rpy' not defined for objects of type '<class 'rpy2.rlike.container.OrdDict'>'
        with localconverter(pandas2ri.converter + ro.default_converter):
            family = getattr(ro.r, self._family)
            kwargs = {}
            if self._knotslocs != KnotLocs.AUTO:
                kwargs["knots"] = pd.DataFrame(
                    _get_knotlocs(
                        self.x,
                        self._n_knots,
                        uniform=False,
                    ),
                    columns=["x"],
                )

            self._model = self._lib.gam(
                ro.Formula(self._formula),
                data=self._design_mat,
                family=family,
                weights=pd.Series(self.w),
                control=self._lib.gam_control(**self._control_kwargs),
                **kwargs,
            )

        return self

    def _get_x_test(self, x_test: Optional[np.ndarray] = None, key_added: str = "_x_test") -> pd.DataFrame:
        newdata = pd.DataFrame(self._check(key_added, x_test), columns=["x"])

        if "offset" in self._design_mat:
            newdata["offset"] = np.mean(self._design_mat["offset"])

        return newdata

    @d.dedent
    def predict(
        self,
        x_test: Optional[np.ndarray] = None,
        key_added: str = "_x_test",
        level: Optional[float] = None,
        **kwargs,
    ) -> np.ndarray:
        """%(base_model_predict.full_desc)s

        This method can also compute the confidence interval.

        Parameters
        ----------
        %(base_model_predict.parameters)s
        level
            Confidence level for confidence interval calculation.
            If :obj:`None`, don't compute the confidence interval. Must be in :math:`[0, 1]`.

        Returns
        -------
        %(base_model_predict.returns)s
        """  # noqa
        import rpy2.robjects as ro
        from rpy2.robjects import pandas2ri
        from rpy2.robjects.conversion import localconverter

        if self.model is None:
            raise RuntimeError("Trying to call an uninitialized model. To initialize it, run `.fit()` first.")
        if self._lib is None:
            raise RuntimeError(f"Unable to run prediction, R package `{self._lib_name!r}` is not imported.")

        if level is not None and not (0 <= level <= 1):
            raise ValueError(f"Expected the confidence level to be in interval `[0, 1]`, found `{level}`.")

        newdata = self._get_x_test(x_test)

        with localconverter(pandas2ri.converter + ro.default_converter):
            res = ro.r.predict(
                self.model,
                newdata=ro.pandas2ri.py2rpy(newdata),
                type="response",
                se=level is not None,
            )

        if level is None:
            self._y_test = np.array(res).squeeze().astype(self._dtype)
        else:
            self._y_test = np.array(res.rx2("fit")).squeeze().astype(self._dtype)
            se = np.array(res.rx2("se.fit")).squeeze().astype(self._dtype)

            level = st.norm.ppf(level + (1 - level) / 2)
            self._conf_int = np.c_[self.y_test - level * se, self.y_test + level * se]

        return self.y_test

    @d.dedent
    def confidence_interval(self, x_test: Optional[np.ndarray] = None, level: float = 0.95, **kwargs) -> np.ndarray:
        """%(base_model_ci.summary)s

        Internally, this method calls :meth:`~cellrank.models.GAMR.predict` to extract the confidence interval,
        if needed.

        Parameters
        ----------
        %(base_model_ci.parameters)s
        level
            Confidence level.

        Returns
        -------
        %(base_model_ci.returns)s
        """  # noqa

        # this is 2x as fast as opposed to calling `robjects.r.predict` again
        # on my PC (Michal):
        #   -`.predict` take ~6.5ms (without se=True, it's ~5.5ms, 20% slowdown)
        #   -`.predict` withouty se=True + `.confidence_interval` take ~12ms
        # this impl. achieves the 2x speedup withouty sacrificing the 20%

        if x_test is not None or level != 0.95 or self._conf_int is None:
            _ = self.predict(x_test=x_test, level=level)

        return self._conf_int

    def _deepcopy_attributes(self, dst: "GAMR") -> None:
        super()._deepcopy_attributes(dst)
        dst._offset = copy.deepcopy(self._offset)
        dst._design_mat = copy.deepcopy(self._design_mat)

    @d.dedent
    def copy(self) -> "GAMR":
        """%(copy)s"""  # noqa
        res = GAMR(
            self.adata,
            self._n_knots,
            distribution=self._family,
            offset=self._offset,
            knotlocs=self._knotslocs,
            perform_import_check=False,
        )
        self._shallowcopy_attributes(res)  # does not copy `prepared`, since we're not copying the arrays

        res._formula = self._formula  # we don't save the basis inside
        res._design_mat = self._design_mat
        res._offset = self._offset

        res._lib = self._lib  # just passing the reference
        res._lib_name = self._lib_name

        res._control_kwargs = self._control_kwargs

        return res

    def __getstate__(self) -> dict:
        return {k: v for k, v in self.__dict__.items() if k != "_lib"}

    def __setstate__(self, state: dict):
        self.__dict__ = state
        self._lib, self._lib_name = _maybe_import_r_lib(self._lib_name, raise_exc=True)


def _maybe_import_r_lib(name: str, raise_exc: bool = False) -> Tuple[Optional[Any], Optional[str]]:
    global _r_lib, _r_lib_name

    if name == _r_lib_name and _r_lib is not None:
        return _r_lib, _r_lib_name

    try:
        from logging import ERROR

        from rpy2.rinterface_lib.callbacks import logger
        from rpy2.robjects import r
        from rpy2.robjects.packages import PackageNotInstalledError, importr

        logger.setLevel(ERROR)
        r["options"](warn=-1)

    except ImportError as e:
        raise ImportError("Unable to import `rpy2`, install it first as `pip install rpy2` version `>=3.3.0`.") from e

    try:
        _r_lib = importr(name)
        _r_lib_name = name
        return _r_lib, _r_lib_name
    except PackageNotInstalledError as e:
        if not raise_exc:
            return None, None
        raise RuntimeError(f"Install R library `{name!r}` first as `install.packages({name!r}).`") from e



import abc
from typing import Any, Union

import numpy as np
import scipy.sparse as sp

from cellrank import logging as logg
from cellrank._utils._utils import _connected, _get_neighs, _symmetric

__all__ = ["ConnectivityMixin", "UnidirectionalMixin", "BidirectionalMixin"]


class ConnectivityMixin:
    """Mixin class that reads kNN connectivities and allows for density normalization."""

    def _read_from_adata(
        self,
        conn_key: str = "connectivities",
        check_connectivity: bool = False,
        check_symmetric: bool = True,
        **kwargs: Any,
    ) -> None:
        super()._read_from_adata(**kwargs)
        self._conn_key = conn_key
        conn = _get_neighs(self.adata, mode="connectivities", key=conn_key)
        self._conn = sp.csr_matrix(conn).astype(np.float64, copy=False)

        if check_connectivity and not _connected(self.connectivities):
            logg.warning("kNN graph is not connected")
        if check_symmetric and not _symmetric(self.connectivities):
            logg.warning("kNN graph is not symmetric")

    def _density_normalize(self, matrix: Union[np.ndarray, sp.spmatrix]) -> Union[np.ndarray, sp.spmatrix]:
        """
        Density normalization by the underlying kNN graph.

        Parameters
        ----------
        matrix
            Matrix to normalize.

        Returns
        -------
        Density normalized matrix.
        """
        logg.debug("Density normalizing the transition matrix")

        q = np.asarray(self.connectivities.sum(axis=0)).squeeze()
        Q = sp.spdiags(1.0 / q, 0, matrix.shape[0], matrix.shape[0])

        return Q @ matrix @ Q

    @property
    def connectivities(self) -> sp.csr_matrix:
        """Underlying connectivity matrix."""
        return self._conn


class UnidirectionalMixin:
    """Mixin specifying that its kernel doesn't is directionless."""

    @property
    def backward(self) -> None:
        """None."""


class BidirectionalMixin(abc.ABC):
    """Mixin specifying that its kernel has forward or backward directions."""

    def __init__(self, *args: Any, backward: bool = False, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if not isinstance(backward, (bool, np.bool_)):
            raise TypeError(f"Expected `backward` to be `bool`, found `{type(backward).__name__}`.")
        self._backward = bool(backward)
        self._init_kwargs["backward"] = backward

    @abc.abstractmethod
    def __invert__(self) -> "BidirectionalMixin":
        pass

    @property
    def backward(self) -> bool:
        """Direction of the process."""
        return self._backward



import abc
import enum
import functools
from typing import Any, Tuple

import numba as nb
import numpy as np

from cellrank._utils._docs import d
from cellrank._utils._enum import ModeEnum
from cellrank.kernels._utils import jit_kwargs, norm, np_mean

__all__ = ["DotProduct", "Cosine", "Correlation", "SimilarityABC"]


class Similarity(ModeEnum):
    DOT_PRODUCT = enum.auto()
    COSINE = enum.auto()
    CORRELATION = enum.auto()


try:
    import jax.numpy as jnp
    from jax import hessian, jit

    @jit
    def _softmax_jax(x: np.ndarray, softmax_scale) -> np.ndarray:
        numerator = x * softmax_scale
        numerator = jnp.exp(numerator - jnp.max(numerator))
        return numerator / jnp.sum(numerator)

    @jit
    def _softmax_masked_jax(x: np.ndarray, mask: np.ndarray, softmax_scale) -> np.ndarray:
        numerator = x * softmax_scale
        numerator = jnp.exp(numerator - jnp.nanmax(numerator))
        numerator = jnp.where(mask, 0, numerator)  # essential

        return numerator / jnp.nansum(numerator)

    @functools.partial(jit, static_argnums=(3, 4))
    def _predict_transition_probabilities_jax(
        X: np.ndarray,
        W: np.ndarray,
        softmax_scale: float = 1.0,
        center_mean: bool = True,
        scale_by_norm: bool = True,
    ) -> np.ndarray:
        if center_mean:
            # pearson correlation, otherwise cosine
            W -= W.mean(axis=1)[:, None]
            X -= X.mean()

        if scale_by_norm:
            denom = jnp.linalg.norm(X) * jnp.linalg.norm(W, axis=1)
            mask = jnp.isclose(denom, 0)
            denom = jnp.where(jnp.isclose(denom, 0), 1, denom)  # essential
            return _softmax_masked_jax(W.dot(X) / denom, mask, softmax_scale)

        return _softmax_jax(W.dot(X), softmax_scale)

    _predict_transition_probabilities_jax_H = hessian(_predict_transition_probabilities_jax)

    _HAS_JAX = True
except ImportError:
    _HAS_JAX = False

    hessian = jit = lambda _: _

    def _predict_transition_probabilities_jax(*_args, **_kwargs):
        raise NotImplementedError("No `jax` installation found.")

    _predict_transition_probabilities_jax_H = _predict_transition_probabilities_jax


@nb.njit(**jit_kwargs)
def _softmax(x: np.ndarray, softmax_scale: float) -> Tuple[np.ndarray, np.ndarray]:
    numerator = x * softmax_scale
    numerator = np.exp(numerator - np.max(numerator))
    return numerator / np.sum(numerator), x


@nb.njit(**jit_kwargs)
def _softmax_masked(x: np.ndarray, mask: np.ndarray, softmax_scale: float) -> Tuple[np.ndarray, np.ndarray]:
    numerator = x * softmax_scale
    numerator = np.exp(numerator - np.nanmax(numerator))
    numerator = np.where(mask, 0, numerator)  # essential

    return numerator / np.nansum(numerator), x


@nb.njit(parallel=False, **jit_kwargs)
def _predict_transition_probabilities_numpy(
    X: np.ndarray,
    W: np.ndarray,
    softmax_scale: float = 1.0,
    center_mean: bool = True,
    scale_by_norm: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    if center_mean:
        # pearson correlation
        W -= np.expand_dims(np_mean(W, axis=1), axis=1)

    if X.shape[0] == 1:
        if center_mean:
            # pearson correlation
            X = X - np.mean(X)

        if scale_by_norm:
            # cosine or pearson correlation
            denom = np.linalg.norm(X) * norm(W, axis=1)
            mask = denom == 0
            denom[mask] = 1
            return _softmax_masked(W.dot(X[0]) / denom, mask, softmax_scale)

        return _softmax(W.dot(X[0]), softmax_scale)

    assert X.shape[0] == W.shape[0], "Wrong shape."

    if center_mean:
        X = X - np.expand_dims(np_mean(X, axis=1), axis=1)

    if scale_by_norm:
        denom = norm(X, axis=1) * norm(W, axis=1)
        mask = denom == 0
        denom[mask] = 1
        return _softmax_masked(
            np.array([np.dot(X[i], W[i]) for i in range(X.shape[0])]) / denom,
            mask,
            softmax_scale,
        )

    return _softmax(np.array([np.dot(X[i], W[i]) for i in range(X.shape[0])]), softmax_scale)


class Hessian(abc.ABC):
    @d.get_full_description(base="hessian")
    @d.get_sections(base="hessian", sections=["Parameters", "Returns"])
    @abc.abstractmethod
    def hessian(self, v: np.ndarray, D: np.ndarray, softmax_scale: float = 1.0) -> np.ndarray:
        """Compute the Hessian.

        Parameters
        ----------
        v
            Array of shape ``(n_genes,)`` containing the velocity vector.
        D
            Array of shape ``(n_neighbors, n_genes)`` corresponding to the transcriptomic displacement of the current
            cell with respect to ist nearest neighbors.
        softmax_scale
            Scaling factor for the softmax function.

        Returns
        -------
        The full Hessian of shape ``(n_neighbors, n_genes, n_genes)`` or only its diagonal of shape
        ``(n_neighbors, n_genes)``.

        Developer notes
        ---------------
        This class should be used in conjunction with any class that implements the ``__call__`` method and should
        compute the Hessian of such function. This it does not need to handle the case of a backward process,
        i.e., when the velocity vector :math:`v` is of shape ``(n_genes, n_neighbors)``.

        If using :mod:`jax` to compute the Hessian, please specify the class attribute ``__use_jax__ = True``.
        """


class SimilarityABC(abc.ABC):
    """Base class for all similarity schemes."""

    @d.get_full_description(base="sim_scheme")
    @d.get_sections(base="sim_scheme", sections=["Parameters", "Returns"])
    @abc.abstractmethod
    def __call__(self, v: np.ndarray, D: np.ndarray, softmax_scale: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:
        """Compute transition probability of a cell to its nearest neighbors using RNA velocity.

        Parameters
        ----------
        v
            Array of shape ``(n_genes,)`` or ``(n_neighbors, n_genes)`` containing the velocity vector(s).
            The second case is used for the backward process.
        D
            Array of shape ``(n_neighbors, n_genes)`` corresponding to the transcriptomic displacement of the current
            cell with respect to ist nearest neighbors.
        softmax_scale
            Scaling factor for the softmax function.

        Returns
        -------
        The probability and unscaled logits arrays of shape ``(n_neighbors,)``.
        """

    @staticmethod
    def create(scheme: Similarity) -> "SimilarityABC":
        if scheme == Similarity.CORRELATION:
            return Correlation()
        if scheme == Similarity.COSINE:
            return Cosine()
        if scheme == Similarity.DOT_PRODUCT:
            return DotProduct()
        raise NotImplementedError(scheme)

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def __str__(self):
        return repr(self)


class SimilarityHessian(SimilarityABC, Hessian):
    """Base class for all similarity schemes as defined in :cite:`li:20`.

    Parameters
    ----------
    center_mean
        Whether to center the velocity vectors and the transcriptomic displacement matrix.
    scale_by_norm
        Whether to scale the velocity vectors) and the transcriptomic displacement matrix by their norms.
    """

    __use_jax__: bool = True

    def __init__(self, center_mean: bool, scale_by_norm: bool):
        self._center_mean = center_mean
        self._scale_by_norm = scale_by_norm

    @d.dedent
    def __call__(self, v: np.ndarray, D: np.ndarray, softmax_scale: float = 1.0) -> Tuple[np.ndarray, np.ndarray]:
        """%(sim_scheme.full_desc)s

        Parameters
        ----------
        %(sim_scheme.parameters)s

        Returns
        -------
        %(sim_scheme.returns)s
        """  # noqa: D400
        return _predict_transition_probabilities_numpy(v, D, softmax_scale, self._center_mean, self._scale_by_norm)

    @d.dedent
    def hessian(self, v: np.ndarray, D: np.ndarray, softmax_scale: float = 1.0) -> np.ndarray:
        """%(hessian.full_desc)s

        Parameters
        ----------
        %(hessian.parameters)s

        Returns
        -------
        %(hessian.returns)s
        """  # noqa: D400
        return _predict_transition_probabilities_jax_H(v, D, softmax_scale, self._center_mean, self._scale_by_norm)

    def __getattribute__(self, name: str) -> Any:
        if name == "hessian" and self.__use_jax__ and not _HAS_JAX:
            raise NotImplementedError("No `jax` installation found.")
        return super().__getattribute__(name)


class DotProduct(SimilarityHessian):
    r"""Dot product scheme as defined in eq. (4.9) :cite:`li:20`.

    .. math::
        v(s_i, s_j) := g(\delta_{i, j}^T v_i)

    where :math:`v_i` is the velocity vector of cell :math:`i`, :math:`\delta_{i, j}` corresponds to the transcriptional
    displacement between cells :math:`i` and :math:`j` and :math:`g` is a softmax function with some scaling parameter.
    """

    def __init__(self):
        super().__init__(center_mean=False, scale_by_norm=False)


class Cosine(SimilarityHessian):
    r"""Cosine similarity scheme as defined in eq. (4.7) :cite:`li:20`.

    .. math::
        v(s_i, s_j) := g(cos(\delta_{i, j}, v_i))

    where :math:`v_i` is the velocity vector of cell :math:`i`, :math:`\delta_{i, j}` corresponds to the transcriptional
    displacement between cells :math:`i` and :math:`j` and :math:`g` is a softmax function with some scaling parameter.
    """

    def __init__(self):
        super().__init__(center_mean=False, scale_by_norm=True)


class Correlation(SimilarityHessian):
    r"""Pearson correlation scheme as defined in eq. (4.8) :cite:`li:20`.

    .. math::
        v(s_i, s_j) := g(corr(\delta_{i, j}, v_i))

    where :math:`v_i` is the velocity vector of cell :math:`i`, :math:`\delta_{i, j}` corresponds to the transcriptional
    displacement between cells :math:`i` and :math:`j` and :math:`g` is a softmax function with some scaling parameter.
    """

    def __init__(self):
        super().__init__(center_mean=True, scale_by_norm=True)



import pathlib
import types
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import numpy as np
from sklearn.preprocessing import StandardScaler

import matplotlib.pyplot as plt
from matplotlib.colors import ListedColormap, is_color_like
from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec

import scanpy as sc
from anndata import AnnData

from cellrank import logging as logg
from cellrank._utils import Lineage
from cellrank._utils._docs import d
from cellrank._utils._enum import DEFAULT_BACKEND, Backend_t
from cellrank._utils._parallelize import _get_n_cores
from cellrank._utils._utils import (
    _check_collection,
    _genesymbols,
    _unique_order_preserving,
    save_fig,
)
from cellrank.pl._utils import (
    _callback_type,
    _create_callbacks,
    _create_models,
    _fit_bulk,
    _get_backend,
    _get_sorted_colors,
    _input_model_type,
    _return_model_type,
    _time_range_type,
)

__all__ = ["cluster_trends"]


@d.dedent
@_genesymbols
def cluster_trends(
    adata: AnnData,
    model: _input_model_type,
    genes: Sequence[str],
    lineage: str,
    time_key: str,
    backward: bool = False,
    time_range: _time_range_type = None,
    clusters: Optional[Sequence[str]] = None,
    n_points: int = 200,
    covariate_key: Optional[Union[str, Sequence[str]]] = None,
    ratio: float = 0.05,
    cmap: Optional[str] = "viridis",
    norm: bool = True,
    recompute: bool = False,
    callback: _callback_type = None,
    ncols: int = 3,
    sharey: Union[str, bool] = False,
    key: Optional[str] = None,
    random_state: Optional[int] = None,
    show_progress_bar: bool = True,
    n_jobs: Optional[int] = 1,
    backend: Backend_t = DEFAULT_BACKEND,
    figsize: Optional[Tuple[float, float]] = None,
    dpi: Optional[int] = None,
    save: Optional[Union[str, pathlib.Path]] = None,
    pca_kwargs: Dict = types.MappingProxyType({"svd_solver": "arpack"}),
    neighbors_kwargs: Dict = types.MappingProxyType({"use_rep": "X"}),
    clustering_kwargs: Dict = types.MappingProxyType({}),
    return_models: bool = False,
    **kwargs: Any,
) -> Optional[_return_model_type]:
    """Cluster and plot gene expression trends within a lineage.

    .. seealso::
        - See :doc:`../../../notebooks/tutorials/estimators/800_gene_trends` on how to
          visualize the gene trends.

    This function is based on *Palantir* :cite:`setty:19`. It can be used to discover modules of genes that drive
    development along a given lineage. Consider running this function on a subset of genes which are potential
    lineage drivers.

    Parameters
    ----------
    %(adata)s
    %(model)s
    %(genes)s
    lineage
        Name of the lineage for which to cluster the genes.
    time_key
        Key in :attr:`~anndata.AnnData.obs` where the pseudotime is stored.
    %(backward)s
    %(time_range)s
    clusters
        Cluster identifiers to plot. If :obj:`None`, all clusters will be considered. Useful when
        plotting previously computed clusters.
    n_points
        Number of points used for prediction.
    covariate_key
        Keys in :attr:`~anndata.AnnData.obs` containing observations to be plotted at the bottom of each plot.
    %(gene_symbols)s
    ratio
        Height ratio of each covariate in ``covariate_key``.
    cmap
        Colormap to use for continuous covariates in ``covariate_key``.
    norm
        Whether to z-normalize each trend to have zero mean, unit variance.
    recompute
        If :obj:`True`, recompute the clustering, otherwise try to find already existing one.
    %(model_callback)s
    ncols
        Number of columns for the plot.
    sharey
        Whether to share y-axis across multiple plots.
    key
        Key in :attr:`~anndata.AnnData.uns` where to save the results.
        If :obj:`None`, it will be saved as ``'lineage_{lineage}_trend'`` .
    random_state
        Random seed for reproducibility.
    %(parallel)s
    %(plotting)s
    pca_kwargs
        Keyword arguments for :func:`~scanpy.pp.pca`.
    neighbors_kwargs
        Keyword arguments for :func:`~scanpy.pp.neighbors`.
    clustering_kwargs
        Keyword arguments for :func:`~scanpy.tl.leiden`.
    %(return_models)s
    kwargs
        Keyword arguments for :meth:`~cellrank.models.BaseModel.prepare`.

    Returns
    -------
    %(plots_or_returns_models)s Also updates :attr:`adata.uns <anndata.AnnData.uns>` with the following:

    - ``key`` or ``'lineage_{lineage}_trend'`` - :class:`~anndata.AnnData` object of
      shape ``(n_genes, n_points)`` containing the clustered genes.
    """

    def plot_cluster(row: int, col: int, cluster: str, sharey_ax: Optional[str] = None) -> Optional[plt.Axes]:
        gss = GridSpecFromSubplotSpec(
            row_delta,
            1,
            subplot_spec=gs[row : row + row_delta, col],
            hspace=0,
            height_ratios=[1.0] + [ratio] * (row_delta - 1),
        )
        ax = fig.add_subplot(gss[0, 0], sharey=sharey_ax)

        data = trends[trends.obs["clusters"] == c].X
        mean, sd = np.mean(data, axis=0), np.std(data, axis=0)

        for i in range(data.shape[0]):
            ax.plot(data[i], color="gray", lw=0.5)

        ax.plot(mean, lw=2, color="black")
        ax.plot(mean - sd, lw=1.5, color="black", linestyle="--")
        ax.plot(mean + sd, lw=1.5, color="black", linestyle="--")
        ax.fill_between(range(len(mean)), mean - sd, mean + sd, color="black", alpha=0.1)

        ax.set_title(f"cluster {cluster}")
        ax.set_xticks([])
        if sharey:
            ax.set_yticks([])
        ax.margins(0)

        if covariate_colors is not None:
            for i, colors in enumerate(covariate_colors):
                ax_clusters = fig.add_subplot(gss[i + 1, 0])
                if is_color_like(colors[0]):  # e.g. categorical
                    cm = ListedColormap(colors, N=len(colors))
                    ax_clusters.imshow(np.arange(cm.N)[None, :], cmap=cm, aspect="auto")
                else:
                    cm = plt.get_cmap(cmap)
                    ax_clusters.imshow(colors[None, :], cmap=cm, aspect="auto")
                ax_clusters.set_xticks([])
                ax_clusters.set_yticks([])

            ax_clusters.set_xticks(np.linspace(0, len(colors), 5))
            ax_clusters.set_xticklabels([f"{v:.3f}" for v in np.linspace(tmin, tmax, 5)])

        return ax if sharey else None

    use_raw = kwargs.get("use_raw", False)
    _ = Lineage.from_adata(adata, backward=backward)[lineage]  # sanity check

    genes = _unique_order_preserving(genes)
    _check_collection(adata, genes, "var_names", use_raw=use_raw)

    if key is None:
        key = f"lineage_{lineage}_trend"

    if recompute or key not in adata.uns:
        kwargs["backward"] = backward
        kwargs["time_key"] = time_key
        kwargs["n_test_points"] = n_points
        models = _create_models(model, genes, [lineage])
        all_models, models, genes, _ = _fit_bulk(
            models,
            _create_callbacks(adata, callback, genes, [lineage], **kwargs),
            genes,
            lineage,
            time_range,
            return_models=True,  # always return (better error messages)
            filter_all_failed=True,
            parallel_kwargs={
                "show_progress_bar": show_progress_bar,
                "n_jobs": _get_n_cores(n_jobs, len(genes)),
                "backend": _get_backend(models, backend),
            },
            **kwargs,
        )

        # `n_genes, n_test_points`
        trends = np.vstack([model[lineage].y_test for model in models.values()]).T

        if norm:
            logg.debug("Normalizing trends")
            _ = StandardScaler(copy=False).fit_transform(trends)

        mod = next(mod for tmp in all_models.values() for mod in tmp.values())
        trends = AnnData(trends.T)
        trends.obs_names = genes
        trends.var["x_test"] = x_test = mod.x_test

        # sanity check
        if trends.n_obs != len(genes):
            raise RuntimeError(f"Expected to find `{len(genes)}` genes, found `{trends.n_obs}`.")
        if trends.n_vars != n_points:
            raise RuntimeError(f"Expected to find `{n_points}` points, found `{trends.n_vars}`.")

        random_state = np.random.RandomState(random_state).randint(2**16)

        pca_kwargs = dict(pca_kwargs)
        pca_kwargs.setdefault("n_comps", min(50, n_points, len(genes)) - 1)
        pca_kwargs.setdefault("random_state", random_state)
        sc.pp.pca(trends, **pca_kwargs)

        neighbors_kwargs = dict(neighbors_kwargs)
        neighbors_kwargs.setdefault("random_state", random_state)
        sc.pp.neighbors(trends, **neighbors_kwargs)

        clustering_kwargs = dict(clustering_kwargs)
        clustering_kwargs["key_added"] = "clusters"
        clustering_kwargs.setdefault("random_state", random_state)
        sc.tl.leiden(trends, **clustering_kwargs)

        logg.info(f"Saving data to `adata.uns[{key!r}]`")
        adata.uns[key] = trends
    else:
        all_models = None
        logg.info(f"Loading data from `adata.uns[{key!r}]`")
        trends = adata.uns[key]
        x_test = trends.var["x_test"]

    if "clusters" not in trends.obs:
        raise KeyError("Unable to find the clustering in `trends.obs['clusters']`.")

    if clusters is None:
        clusters = trends.obs["clusters"].cat.categories
    for c in clusters:
        if c not in trends.obs["clusters"].cat.categories:
            raise ValueError(
                f"Invalid cluster name `{c!r}`. " f"Valid options are `{list(trends.obs['clusters'].cat.categories)}`."
            )

    nrows = int(np.ceil(len(clusters) / ncols))
    fig = plt.figure(
        dpi=dpi,
        figsize=(ncols * 10, nrows * 10) if figsize is None else figsize,
        tight_layout=True,
    )

    if covariate_key is None:
        covariate_colors, row_delta = None, 1
    else:
        tmin, tmax = np.min(x_test), np.max(x_test)
        covariate_colors = _get_sorted_colors(
            adata,
            covariate_key,
            time_key,
            tmin=tmin,
            tmax=tmax,
        )
        row_delta = len(covariate_colors) + 1
    gs = GridSpec(nrows=nrows * row_delta, ncols=ncols, figure=fig)

    row, sharey_ax = -row_delta, None
    for i, c in enumerate(clusters):
        if i % ncols == 0:
            row += row_delta
        sharey_ax = plot_cluster(row, i % ncols, c, sharey_ax=sharey_ax)

    if save is not None:
        save_fig(fig, save)

    if return_models:
        return all_models
